/*
dsound.cpp
directsound functions ...

Copyright (C) 1995-2005  Everett Bell

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "globals.h"

#include "wassert.h"
#include "wave.h"


#include <winnls.h>

#include <direct.h>


/*

int WaveOpenFile(
	char *pszFileName,                              // (IN)
	HMMIO *phmmioIn,                                // (OUT)
	WAVEFORMATEX **ppwfxInfo,                       // (OUT)
	MMCKINFO *pckInRIFF                             // (OUT)
			);

int WaveStartDataRead(
	HMMIO *phmmioIn,
	MMCKINFO *pckIn,
	MMCKINFO *pckInRIFF
	);

int WaveReadFile(
	HMMIO hmmioIn,                          // IN
	UINT cbRead,                            // IN           
	BYTE *pbDest,                           // IN
	MMCKINFO *pckIn,                        // IN.
	UINT *cbActualRead                      // OUT.
	);

int WaveCloseReadFile(
	HMMIO *phmmio,                                  // IN
	WAVEFORMATEX **ppwfxSrc                 // IN
	);

*/





BOOL InitDSound(HWND hwnd, GUID *pguid)
{
   // Create DirectSound
	HRESULT hr;

 
    if FAILED(DirectSoundCreate(pguid, &lpds, NULL))
	{
sprintf( debugstr, "Failed DirectSoundCreate()!\n" );
WRITETODEBUGFILE;
        return FALSE;
	}
 
    // Set co-op level
 
    if FAILED(IDirectSound_SetCooperativeLevel(
            lpds, hwnd, DSSCL_PRIORITY))
	{
sprintf( debugstr, "Failed IDirectSound_SetCooperativeLevel()!\n" );
WRITETODEBUGFILE;
        return FALSE;
	}
 
//You have set the priority cooperative level in order to be able to set the format of the primary buffer. If you don't change the default format, output will be in the 8-bit, 22 kHz format regardless of the format of the input. Setting the primary buffer to a higher format can do no harm, because even if the secondary buffers are in a lower format, the samples will be converted automatically by DirectSound. Note also that there's no danger of the call to IDirectSoundBuffer::SetFormat failing because the hardware doesn't support the higher format. DirectSound will simply set the closest available format.

//To set the format of the primary buffer, you first describe it in the global DSBUFFERDESC structure, then pass this description to the IDirectSound::CreateSoundBuffer method.

    // Obtain primary buffer
 
    ZeroMemory(&dsbdesc, sizeof(DSBUFFERDESC));
    dsbdesc.dwSize = sizeof(DSBUFFERDESC);
    dsbdesc.dwFlags = DSBCAPS_PRIMARYBUFFER;
    if FAILED(lpds->CreateSoundBuffer(&dsbdesc, &lpdsbPrimary, NULL))
	{
sprintf( debugstr, "Failed lpds->CreateSoundBuffer()!\n" );
WRITETODEBUGFILE;
        return FALSE;
	}
 
//Once you have a primary buffer object, you describe the desired wave format and pass the description to the IDirectSoundBuffer::SetFormat method:

    // Set primary buffer format
 
    WAVEFORMATEX wfx;
    memset(&wfx, 0, sizeof(WAVEFORMATEX)); 
    wfx.wFormatTag = WAVE_FORMAT_PCM; 
    wfx.nChannels = 2; 
    wfx.nSamplesPerSec = 44100; 
    wfx.wBitsPerSample = 16; 
    wfx.nBlockAlign = wfx.wBitsPerSample / 8 * wfx.nChannels;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
 
    hr = lpdsbPrimary->SetFormat(&wfx); 
 
    return TRUE;
}  // InitDSound()
 
 










BOOL SetupStreamBuffer(LPSTR lpzFileName)
{
    // Close any open file and release interfaces
 
    WaveCloseReadFile(&hmmio, &pwfx);  // Function in Wave.c
 
    if (lpdsNotify != NULL)
    {
        lpdsNotify->Release();
        lpdsNotify = NULL;
    }
 
    if (lpdsb != NULL)
    {
        lpdsb->Release();
        lpdsb = NULL;
    }
 
//The function now opens a wave file, gets the format, and advances the file pointer to the beginning
//	of the sound data. It does this by using two functions in Wave.c.

    if (WaveOpenFile(lpzFileName, &hmmio, &pwfx, 
            &mmckinfoParent) != 0)
        return FALSE;
    if (WaveStartDataRead(&hmmio, &mmckinfoData, 
            &mmckinfoParent) != 0)
        return FALSE;
 
//The WaveOpenFile function initializes three of the global variables declared at the beginning of
//	the tutorial: a file handle, a pointer to a WAVEFORMATEX structure, and the MMCKINFO structure
//	for the parent chunk (that is, information about the file as a whole).

//The file handle and chunk information are then passed to the WaveStartDataRead function, which 
//	returns information about the data chunk in mmckinfoData. You would be interested in this structure 
//	if you were creating a static buffer just big enough to accommodate all the data bytes. In this 
//	tutorial, though, you're creating a streaming buffer, so you don't need to know the size of the data chunk.

 
    memset(&dsbdesc, 0, sizeof(DSBUFFERDESC)); 
    dsbdesc.dwSize = sizeof(DSBUFFERDESC); 
    dsbdesc.dwFlags = 
            DSBCAPS_GETCURRENTPOSITION2   // Always a good idea
            | DSBCAPS_GLOBALFOCUS         // Allows background playing
            | DSBCAPS_CTRLPOSITIONNOTIFY; // Needed for notification
 
    // The size of the buffer is arbitrary, but should be at least
    // two seconds, to keep data writes well ahead of the play
    // position.
 
    dsbdesc.dwBufferBytes = pwfx->nAvgBytesPerSec * 2;  
    dsbdesc.lpwfxFormat = pwfx; 
 
    if FAILED(IDirectSound_CreateSoundBuffer(
            lpds, &dsbdesc, &lpdsb, NULL))
    {
        WaveCloseReadFile(&hmmio, &pwfx);
        return FALSE; 
    }
 
 
    for (int i = 0; i < NUMEVENTS; i++)
    {
        rghEvent[i] = CreateEvent(NULL, FALSE, FALSE, NULL);
        if (NULL == rghEvent[i]) return FALSE;
    }
 
//Next, you initialize the array of DSBPOSITIONNOTIFY structures, each of which associates a position in 
//the buffer with an event handle:

    rgdsbpn[0].dwOffset = 0;
    rgdsbpn[0].hEventNotify = rghEvent[0];
    rgdsbpn[1].dwOffset = (dsbdesc.dwBufferBytes/2);
    rgdsbpn[1].hEventNotify = rghEvent[1];
 
//Then you get the IDirectSoundNotify interface from the secondary buffer and pass the DSBPOSITIONNOTIFY 
//array to the SetNotificationPositions method:

    if FAILED(IDirectSoundBuffer_QueryInterface(lpdsb, 
            IID_IDirectSoundNotify, (VOID **)&lpdsNotify))
        return FALSE; 
 
    if FAILED(IDirectSoundNotify_SetNotificationPositions(
             lpdsNotify, NUMEVENTS, rgdsbpn))
    {
        IDirectSoundNotify_Release(lpdsNotify);
        return FALSE;
    }
 
//That completes the setup of the streaming buffer. Because you've already opened the wave file and are
//ready to start streaming data, you can set the buffer in motion here. You want it to continue running
//till the complete sound has been played, so you set the DSBPLAY_LOOPING flag.

    IDirectSoundBuffer_Play(lpdsb, 0, 0, DSBPLAY_LOOPING);
 
    return TRUE;
} // end of SetupStreamBuffer()
 
 




BOOL StreamToBuffer(DWORD dwPos)
{
    LONG            lNumToWrite;
    DWORD           dwStartOfs;
    VOID            *lpvPtr1, *lpvPtr2;
    DWORD           dwBytes1, dwBytes2;
    UINT            cbBytesRead;
    static DWORD    dwStopNextTime = 0xFFFF;
 
    if (dwStopNextTime == dwPos)   // All data has been played
    {
        lpdsb->Stop();
        dwStopNextTime = 0xFFFF;
        return TRUE;
    }
 
    if (dwStopNextTime != 0xFFFF)  // No more to stream, but keep
                                 // playing to end of data
        return TRUE;
 
//The dwStopNextTime variable is a flag to indicate that the end of the file has been reached;
//this is set later in the function. If it is nonzero, there is no more data to be streamed. If 
//the value of dwStopNextTime equals the index of the notification being handled, then you know
//that the current play position has returned to where it was when the end of the file was reached;
//that is, the last segment of data you copied to the buffer has been played. In this case, it's
//time to stop the buffer so it doesn't keep playing old data.

//The next part of the StreamToBuffer function determines the offset within the buffer where you
//will start copying the new data. Although the buffer in this tutorial has only two notification
//positions, the code is designed to work with any number. Remember, dwPos is the index of the
//notification position that has just been passed by the current play position. You are going to
//write to the segment of the buffer that starts at the previous notification position. 

    if (dwPos == 0)
        dwStartOfs = rgdsbpn[NUMEVENTS - 1].dwOffset;
    else
        dwStartOfs = rgdsbpn[dwPos-1].dwOffset;
 
//Now you determine the size of this segment of the buffer:

    lNumToWrite = (LONG) rgdsbpn[dwPos].dwOffset - dwStartOfs;
    if (lNumToWrite < 0) lNumToWrite += dsbdesc.dwBufferBytes;
 
//You now have all the information you need to lock the buffer in preparation for the data write.

    IDirectSoundBuffer_Lock(lpdsb,
                     dwStartOfs,       // Offset of lock start
                     lNumToWrite,      // Number of bytes to lock
                     &lpvPtr1,         // Address of lock start
                     &dwBytes1,        // Count of bytes locked
                     &lpvPtr2,         // Address of wrap around
                     &dwBytes2,        // Count of wrap around bytes
                     0);               // Flags
 
//In this example the lock will never wrap around, so you need to do only a single data copy from
//the file. You get the data by calling the WaveReadFile function in Wave.c:

    WaveReadFile(hmmio,                // File handle
                 dwBytes1,             // Number of bytes to get
                 (BYTE *) lpvPtr1,     // Destination   
                 &mmckinfoData,        // File chunk info 
                 &cbBytesRead);        // Actual bytes read
 
//Now determine if the end of the file has been reached. If it has, close it, write silence to the
//rest of this segment (because it will play all the way through before the buffer is stopped), and
//set the dwStopNextTime flag.

    if (cbBytesRead < dwBytes1)        // Reached end of file
    {
        WaveCloseReadFile(&hmmio, &pwfx);
        FillMemory((PBYTE)lpvPtr1 + cbBytesRead,
                dwBytes1 - cbBytesRead, 
                (dsbdesc.lpwfxFormat->wBitsPerSample==8) ? 128 : 0);
        dwStopNextTime = dwPos;
    }
 
//Finally, unlock the buffer and return to the message loop:

    IDirectSoundBuffer_Unlock(lpdsb, 
            lpvPtr1, dwBytes1, lpvPtr2, dwBytes2);
    
    return TRUE;
} // end StreamToBuffer()
 





BOOL LoadSound( dword soundno, char *lpszfilename )
{
	WAVEFORMATEX  *pwfx;          // Wave format info
	HMMIO         hmmio;          // File handle
	MMCKINFO      mmckinfoData;   // Chunk info
	MMCKINFO      mmckinfoParent; // Parent chunk info

	LPVOID lpvAudio1;
	DWORD  dwBytes1;

	HRESULT hr;

	UINT cbBytesRead;
	
//The first step in reading a wave file is to call the WaveOpenFile function. This gets a handle to the file, verifies that it is in RIFF format, and gets information about the wave format. The parameters are the filename and the addresses of three of the variables you have declared:

	if (WaveOpenFile(lpszfilename, &hmmio, &pwfx, &mmckinfoParent) != 0)
	{
sprintf( debugstr, "WaveOpenFile() Failed!" );
WRITETODEBUGFILE;
		return FALSE;
	}
 
//Note that the wrapper functions all return zero if successful.

//The next step is to call the WaveStartDataRead function, causing the file pointer to descend to the data chunk. This function also fills in the MMCKINFO structure for the data chunk, so that you know how much data is available:

	if (WaveStartDataRead(&hmmio, &mmckinfoData, &mmckinfoParent) != 0)
    {
sprintf( debugstr, "WaveStartDataRead() Failed!" );
WRITETODEBUGFILE;
		return FALSE;
    }
 
//The application can now begin copying data from the file to a secondary sound buffer. Normally you
//	don't create the sound buffer until you have obtained the size of the data chunk and the format of
//	the wave. The following code creates a static buffer just large enough to hold all the data in the file. 
 
	memset(&dsbufferdesc[soundno], 0, sizeof(DSBUFFERDESC)); 
	dsbufferdesc[soundno].dwSize = sizeof(DSBUFFERDESC); 
	dsbufferdesc[soundno].dwFlags = DSBCAPS_STATIC; 
	dsbufferdesc[soundno].dwBufferBytes = mmckinfoData.cksize;  
	dsbufferdesc[soundno].lpwfxFormat = pwfx; 

	hr = lpds->CreateSoundBuffer(&dsbufferdesc[soundno], &lpdsbuffer[soundno], NULL);
 
	if ( hr != DS_OK )
	{
	    WaveCloseReadFile(&hmmio, &pwfx);

sprintf( debugstr, "CreateSoundBuffer() Failed!" );
WRITETODEBUGFILE;
switch( hr )
{
case DSERR_ALLOCATED:		sprintf( debugstr, " DSERR_ALLOCATED" );	break;
case DSERR_BADFORMAT:		sprintf( debugstr, " DSERR_BADFORMAT" );	break;
case DSERR_INVALIDPARAM:	sprintf( debugstr, " DSERR_INVALIDPARAM" );	break;
case DSERR_NOAGGREGATION:	sprintf( debugstr, " DSERR_NOAGGREGATION" );	break;
case DSERR_OUTOFMEMORY:		sprintf( debugstr, " DSERR_OUTOFMEMORY" );	break;
case DSERR_UNINITIALIZED:	sprintf( debugstr, " DSERR_UNINITIALIZED" );	break;
case DSERR_UNSUPPORTED:		sprintf( debugstr, " DSERR_UNSUPPORTED" );	break;
} // end switch on error code

WRITETODEBUGFILE;

		return FALSE; 
	}
 
//Because in this case the application is not streaming the data but simply filling a static buffer,
//	the entire buffer is locked from the beginning. There is no wraparound, so only a single pointer
//	and byte count are required.

 
	if FAILED(lpdsbuffer[soundno]->Lock(
        0,              // Offset of lock start
        0,              // Size of lock; ignored in this case
        &lpvAudio1,     // Address of lock start
        &dwBytes1,      // Number of bytes locked
        NULL,           // Wraparound start; not used
        NULL,           // Wraparound size; not used
        DSBLOCK_ENTIREBUFFER))  // Flag
	{
	    // Error handling
	    WaveCloseReadFile(&hmmio, &pwfx);

sprintf( debugstr, "SoundBuffer->Lock() Failed!" );
WRITETODEBUGFILE;

		return FALSE;
	}
 
// The WaveReadFile function in Wave.c copies the data from the file to the buffer pointer and returns zero if successful. 

 
	if (WaveReadFile(
        hmmio,              // file handle
        dwBytes1,           // no. of bytes to read
        (BYTE *) lpvAudio1, // destination
        &mmckinfoData,      // file chunk info 
        &cbBytesRead))      // actual no. of bytes read
	{
	    // Handle failure on non-zero return
	    WaveCloseReadFile(&hmmio, &pwfx);

sprintf( debugstr, "WaveReadFile() Failed!" );
WRITETODEBUGFILE;

		return FALSE;
	}
 
//Finally, the application unlocks the buffer and closes the wave file:

	lpdsbuffer[soundno]->Unlock(lpvAudio1, dwBytes1, NULL, 0);
	WaveCloseReadFile(&hmmio, &pwfx);

	return TRUE;
 
} // end LoadSound









 


void DSExit(void)
{
    WaveCloseReadFile(&hmmio, &pwfx);
    if (lpdsNotify)
        lpdsNotify->Release();
    if (lpds)
        lpds->Release(); 
}
 
 












// TODO -- Move to DMusic.cpp !!

IDirectMusicPerformance* CreatePerformance(void)
{
    IDirectMusicPerformance* pPerf;
	int result;


	pPerf = NULL;
	
	result = CoCreateInstance(
            CLSID_DirectMusicPerformance,            
			NULL,
            CLSCTX_INPROC,
			IID_IDirectMusicPerformance,
			(void **)&pPerf
			);

	switch( result )
	{
	default:
		sprintf( tempstr, "pPerf = 0x%.8x", pPerf );
		cleanup( NOSOUND, tempstr );

		pPerf = NULL;
		break;

	case E_INVALIDARG:
		cleanup( NOSOUND, "E_INVALIDARG" );
		break;

	case S_OK:
//		cleanup( NOSOUND, "S_OK" );
		break;

	case REGDB_E_CLASSNOTREG:
		cleanup( NOSOUND, "REGDB_E_CLASSNOTREG" );
		break;

	case CLASS_E_NOAGGREGATION:
		cleanup( NOSOUND, "CLASS_E_NOAGGREGATION" );
		break;

    }


	
    return pPerf;

} // end CreatePerformance









IDirectMusicLoader* CreateLoader(void)
{    
	IDirectMusicLoader* pLoader; 

    if (FAILED(CoCreateInstance(            
		CLSID_DirectMusicLoader,
		NULL,            
		CLSCTX_INPROC,             
		IID_IDirectMusicLoader,
		(void**)&pLoader        )))
	{        
		pLoader = NULL;    
	}

    return pLoader;

} // end CreateLoader








IDirectMusicSegment* LoadSegment( IDirectMusicLoader* pLoader, char *filename )
{
    DMUS_OBJECTDESC ObjDesc; 
    IDirectMusicObject* pObjectSeg = NULL;
    IDirectMusicSegment* pSegment = NULL;
    char szDir[_MAX_PATH];
    WCHAR wszDir[_MAX_PATH];
    WCHAR wfilename[_MAX_PATH];

	HRESULT result;
 
    if(_getcwd( szDir, _MAX_PATH ) == NULL) { return NULL; }
//	strcpy( szDir, homedir );
    
	MULTI_TO_WIDE(wszDir, szDir);

	MULTI_TO_WIDE(wfilename, filename);


/*
	if( (MultiByteToWideChar(  
		CP_ACP, // UINT CodePage,         // code page
		NULL, // DWORD dwFlags,         // character-type options
		szDir, // LPCSTR lpMultiByteStr, // address of string to map
		-1, // int cchMultiByte,      // number of bytes in string
		wszDir, // LPWSTR lpWideCharStr,  // address of wide-character buffer
		_MAX_PATH // int cchWideChar        // size of buffer);
		) == 0) )
	{
		return NULL;
	} // end if failed
*/




    result = pLoader->SetSearchDirectory(GUID_DirectMusicAllTypes,
            wszDir, FALSE );

	switch( result )
	{
	default:
		cleanup( NOSOUND, "SetSearchDirectory() failed." );

	case S_OK:
	case S_FALSE:	// Search Directory already set ...
		break;

	} // end switch on result

	
 
// You then describe the object to be loaded, in a DMUS_OBJECTDESC structure:

    ObjDesc.guidClass = CLSID_DirectMusicSegment;
    ObjDesc.dwSize = sizeof(DMUS_OBJECTDESC);
    wcscpy( ObjDesc.wszFileName, wfilename );
    ObjDesc.dwValidData = DMUS_OBJ_CLASS | DMUS_OBJ_FILENAME;
 
// Now load the object, query it for the IDirectMusicSegment interface, and release it. Note that loading the object also creates the segment, initializes the tracks, and does everything else necessary to get the MIDI data ready for playback.

    result = pLoader->GetObject( &ObjDesc, IID_IDirectMusicSegment, (void**) &pSegment );

	switch( result )
	{
	case DMUS_S_PARTIALLOAD:
	case S_OK:
		break;

	default:
		cleanup( NOFILE, filename );

	} // end switch on result


    result = pSegment->SetParam(GUID_StandardMIDIFile, -1, 0, 0, (void*)g_pPerf);
	switch( result )
	{
	case S_OK:
		break;
	default:
		cleanup( NOSOUND, "SetParam() failed." );
	} // end switch on result

    result = pSegment->SetParam(GUID_Download, -1, 0, 0, (void*)g_pPerf);
	switch( result )
	{
	case S_OK:
		break;
	default:
		cleanup( NOSOUND, "SetParam() failed." );
	} // end switch on result

    return pSegment;
 
} // End of LoadSegment()
