/*
define.h
constants and objects for rush hour ...

Copyright (C) 1995-2005  Everett Bell

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __MYDEFINE
#define __MYDEFINE

#include <windows.h>

#include "resource.h"




#define DEBUG
//#define EDITORS
//#define SHAREWARE




// Video Surfaces ...
#define DDS_CURSORS			0
#define DDS_INTERFACE		1
#define DDS_CITYMAP			2

// Secondary Surfaces ...
#define DDS_SECTIONS		3
#define DDS_GROUND			4

// Fonts ...
#define DDS_EUROFONT		5
#define DDS_EUROFONTDARK	6

// Misc ...
#define DDS_WARNINGLEVEL	7

#define DDS_TOPBAR			8
#define DDS_BUTTON1			9

// Cars ...
#define DDS_CARS			10

#define DDS_MMNOLIGHT		11
#define DDS_MMREDLIGHT		12
#define DDS_MMYELLOWLIGHT	13
#define DDS_MMGREENLIGHT	14

#define DDS_CHOOSEINTERFACE	15


#define NUMSURFACES		16
#define NUMOFFONTS		2


// Language offsets ...
#define ENGLISH			0
#define GERMAN			100
#define SPANISH			200
#define FRENCH			300

// Font types ...

#define EURO			0
#define EURODARK		1


#define byte	unsigned char
#define word	unsigned short int
#define dword	unsigned int


#define SCREENWIDTH		800
#define SCREENHEIGHT	600
#define COLORDEPTH		16
#define VSCREENWIDTH	800
#define VSCREENHEIGHT	600
#define VSOFFSET		0

#define VSLEFT		0
#define VSTOP		0
#define VSRIGHT		800
#define VSBOTTOM	600



// Game Screens ...
#define GS_MAINMENU		1
#define GS_OKBOX		2
#define GS_CONFIRM		3
#define GS_SELECTLIST	4
#define GS_EDITMAP		5
#define GS_GAME			6
#define GS_MAKEPATH		7
#define GS_CHOOSEFILE	8
#define GS_CREDITS		9
#define GS_WAITFORINPUT	10


// Exit Errors ...

#define OK              0
#define OUTOFMEM        1
#define NOVIDEO         2
#define NOFILE          3
#define NOSOUND         4
#define NOMOUSE         5
#define MISC            6




// Cursor Values ...

#define CURSORANIMSPEED	4

// Cursor modes ...

#define CURSOROFF       0
#define MENUCURSOR		1
#define SWITCHCURSOR	2


#define SPRITEONCE	0	// Used in showspriteonce()

// #define MENUBORDER  56



// Button function macros ...

#define BUTTONRETURN          0
#define NOTHING               1
#define SLIDER                2
#define CHECKBOX              3
#define RADIO1                4
#define RADIO2                5
#define RADIO3                6
#define RADIO4                7
#define RADIO5                8
#define RADIO6                9
#define RADIO7                10
#define RADIO8                11

// Offset into button BMP of depressed version
// (should change for different height buttons) ...
#define BUTTONOFF       512L*24L


// 16bit stuff ...
// Pixel formats ...
#define PF_555		1
#define PF_565		2
#define PF_OTHER	3






// Function MACROS ...

/*
* For 16 bit only!
*

// Sets colors ... IN 555 format!
#define RGB16(r,g,b) (word)((r<<10)|(g<<5)|(b))
#define GETR16(color) (((word)(color & 0x7C00))>>10)
#define GETG16(color) (((word)(color & 0x03E0))>>5)
#define GETB16(color) (word)(color & 0x001F)

#define SETTEXTCOLOR(r,g,b) { { textr = r; if( pixelformat==PF_555 ) textg = g; else { textg = g<<1; }; textb = b; } }

#define SETSELECTCOLOR(r,g,b) { selectcolorr = r; selectcolorg = g; selectcolorb = b; }
#define SETSELECTCOLORBASE(r,g,b) { selectcolorbaser = r; selectcolorbaseg = g; selectcolorbaseb = b; }

*/

#define times768(x) (((x)<<9)+((x)<<8))
#define times640(x) (((x)<<9)+((x)<<7))
#define times800(x) (((x)<<9)+((x)<<8) + ((x)<<5))

#define times20(x)  (((x)<<4)+((x)<<2))
#define times10(x)  (((x)<<3)+((x)<<1))



// Locks Virtual Screen ...
// Use indrawscreen to wrap numerous functions that directly access the vscreen ...
#define LOCKVS if( !indrawscreen ) lockvs();
#define UNLOCKVS if( !indrawscreen ) unlockvs();


// Debugging ...
#define TIMESTAMP			{ sprintf( debugstr2, "%.8d - GS: %.2d,  ", tickcount, gamescreen ); WriteFile( debugfh, debugstr2, strlen(debugstr2), &numofbytesread, NULL ); }

#define WRITETODEBUGFILE	{ TIMESTAMP; WriteFile( debugfh, debugstr, strlen(debugstr), &numofbytesread, NULL ); }




#define MULTI_TO_WIDE( x, y )  MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, y, -1, x, _MAX_PATH );



#define pixel2rgb( srcpixel )							\
	rgb16.r = ((srcpixel & rmask) >> rbitshift);		\
	rgb16.g = ((srcpixel & gmask) >> gbitshift);		\
	rgb16.b = ((srcpixel & bmask) >> bbitshift);		


#define rgb2pixel	pixel = (rgb16.r << rbitshift) | (rgb16.g << gbitshift) | (rgb16.b << bbitshift);







// Structures ...

typedef struct pcxheader_typ {
	char manufacturer;
    char version;
    char encoding;
    char bitsperpixel;
    short int x, y;                       // Upper left corner
    short int width, height;              // Size of image
    short int horzres;
    short int vertres;
    char egapalette[48];
    char reserved;
    char numcolorplanes;
    short int bytesperline;
    short int palettetype;
    char padding[58];

} PCXHEADER;


typedef struct sprite_typ
{
	word *picptr;
    word *bgptr;
    short int height, width;
    short int x, y;
    byte visible;

    byte beingused;

	byte blendamt;

} SPRITE;

typedef struct animsprite_typ
{
	byte *picptr[32];
    byte *bgptr;
    short int height, width;
    short int x, y;
    byte visible;
    byte animating;
    byte frame;

    byte animspeed;
    unsigned int clockmax;

    byte numframes;

    byte beingused;

	byte blendamt;

} ANIMSPRITE;

typedef struct letter_typ
{
	short int offset;
	byte width;
} LETTER;

typedef struct myfont_typ
{
	LETTER letter[96];
	byte *picptr;
	byte height;
	int width;
	int surfaceno;
} MYFONT;

typedef struct button_typ
{
    int x1, y1;
    int width;
    int height;
    char pushed;
    char name[32];
    word color;
    char function;          // Macro for what button does when clicked
    int variable;           // Variable to be applied with function (i.e. add)
    int value;
    char *memptr;           // Optional pointer to function call
    char glowing;           // Flag for when mouse is over button
    int offset;
    byte page;

} BUTTON;


typedef	struct rgb16_typ
{
	word r;
	word g;
	word b;
} RGB16;



typedef struct selectoption_typ
{
	char text[64];
	byte disabled;
} SELECTOPTION;



//
// Rush Hour specific structures ...
///////////////////////////////////////////////////////////////////////////////////////////////////////

#define CITYFILE	0
#define CAMPFILE	1

#define AMBULANCE		16


#define GRIDSIZE			20
#define GRIDWIDTH			30
#define NUMOFCITYSECTIONS	112



// Directions ...
#define N	0
#define NNE	1
#define NE	2
#define ENE	3
#define E	4
#define ESE	5
#define SE	6
#define SSE	7
#define S	8
#define SSW	9
#define SW	10
#define	WSW	11
#define W	12
#define WNW	13
#define NW	14
#define NNW	15


#define GAMEWON		2
#define NEWCITY		3



// Light settings ...
#define HORZRED			1
#define HORZYELLOW		2
#define VERTRED			3
#define VERTYELLOW		4
#define COPATLIGHT		5
#define WRECKATLIGHT	6


typedef struct citysection_typ
{
	int graphicx, graphicy;
	byte xsize, ysize;

	BOOL isintersection;

} CITYSECTION;



#define MAXNUMOFCARS	64

typedef struct car_typ
{
	int carid;

	int graphicx, graphicy;
	word homex, homey; // graphic home coordinates

	byte facing;
	byte path[256];
	byte pathno;

	byte citymapx, citymapy;
	byte carmapx, carmapy;
	byte screenx, screeny;

	char xoffset, yoffset; // offset from space

	byte xsize, ysize;

	BOOL onscreen;

	DWORD nextmoveat;
	byte speed;

	byte spaceshift;

	byte stopped;

	void *nextcar;
	void *prevcar;

} CAR;


typedef struct possiblepath_typ
{
	byte startx, starty;
	byte path[256];
	byte pathno;

	BOOL isused;

} POSSIBLEPATH;



typedef struct city_typ
{
	char name[32];
	int numofpaths;
} CITY;


// End one time only definitions ...
#endif




// Function Declarations


// Dialog Box functions ...
BOOL CALLBACK filedialogfunc( HWND hdwnd, UINT message, WPARAM wParam, LPARAM lParam );


int selectlist( void );
void addtoselectlist( char *text, byte disabled );

void waitforinput( int inputtype );
void systemmessage( char *textptr );


BOOL lockvs( void );
BOOL unlockvs( void );


// Message Handlers for various gamescreens ...
int mainmenumessagehandler( UINT message, WPARAM wParam, LPARAM lParam );
int okboxmessagehandler( UINT message, WPARAM wParam, LPARAM lParam );
int confirmmessagehandler( UINT message, WPARAM wParam, LPARAM lParam );
int choosefilemessagehandler( UINT message, WPARAM wParam, LPARAM lParam );

int waitmessagehandler( UINT message, WPARAM wParam, LPARAM lParam );
int selectlistmessagehandler( UINT message, WPARAM wParam, LPARAM lParam );

int gamemessagehandler( UINT message, WPARAM wParam, LPARAM lParam );
int editmapmessagehandler( UINT message, WPARAM wParam, LPARAM lParam );

int makepathmessagehandler( UINT message, WPARAM wParam, LPARAM lParam );

int creditsmessagehandler( UINT message, WPARAM wParam, LPARAM lParam );

BOOL refreshscreen( void );

byte pushgamescreen( byte gs );
byte popgamescreen( void );


long FAR PASCAL WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );


void showpixelformat( void );

void finiObjects( void );


word *loadandunpackbmp( char *tempfilename );


void initspritefrombmp( byte spriteno, char *filename );

// IDirectDrawSurface4 *DDLoadBitmapSYS(IDirectDraw4 *pdd, LPCSTR szBitmap, int dx, int dy);

void ShowDDError(char *where, HRESULT hErr);

HRESULT restoreAll( void );

void clearscreen( void );
void clearvscreen( void );

void screenshot( void );

void fadeinandout( char *bmpfilename );
void fadein( char *bmpfilename );
void fadeout( char *bmpfilename );


void CALLBACK mytimer(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2);



/////////////////////////
// OLD PROTOTYPES !!!! //
/////////////////////////

byte confirm( char *text );
void okbox( char *textptr );


int loadfile( char *path, char *par );

void mousemessage( char *textptr );


void drawwindow( int x, int y, int width, int height );

void wait( void );
void waitfor( int numticks );

void screenshot( void );


// Sprite functions ...

void initsprite( byte spriteno, short int width, short int height );
void deletesprite( byte spriteno );
void showsprite( byte spriteno );
void showspriteonce( char *tempfilename, int x, int y );
void hidesprite( byte spriteno );
void animatesprites( byte numsprites );
void initanimsprite( byte spriteno, short int width, short int height, byte numframes,
                      byte speed, byte animating );
void deleteanimsprite( byte spriteno );
void showanimsprite( byte spriteno );
void hideanimsprite( byte spriteno );

void scalesprite( byte srcsprite, byte destsprite, int xsize, int ysize );



void cleanup( char errortype, char *text );


void initscreen( void );
void initgraphics( void );
void drawscreen( void );

void displayscreen( void );


void showcursor( void );
byte *unpackpcx( char *filename );
void showbitmap( char *tempfilename, int x, int y );


void setpalette(char *Palette);
void setpalfrompcx( char *filename );


void inittext( void );
int getlength( char *text );
void showchar( char c, int x, int y, char recessed );
void showtext( char *text, int x, int y );
void centertext( char *text, int y, int min, int max );
void showtextjr( char *text, int x, int y );

void showtextfile( char *name );

void mouse_setpos( int x, int y );
void updatemouse( void );
void showcursor( void );
void hidecursor( void );
void setcursormode( byte mode );
void animatecursor( void );   // Set up as an interrupt if not too slow

/*
void initgraphics( void );
void drawbutton( char buttonno );
byte popupbox( void );
*/



void drawbutton( char buttonno );


// DirectSound ...

#define NUMEVENTS	8
#define NUMSOUNDS	16

#define SND_1		1
#define SND_2		2
#define SND_3		3
#define SND_4		4
#define SND_5		5
#define SND_6		6
#define SND_7		7
#define SND_8		8
#define SND_9		9
#define SND_10		10
#define SND_11		11
#define SND_12		12
#define SND_13		13
#define SND_14		14
#define SND_15		15

#define SND_CLICK		1
#define SND_COPAPPEARS	2
#define SND_COPSIREN	3
#define SND_COPURGENT	4
#define SND_CARHORN		5
#define SND_CARHORN2	6

#define SND_WINGAME		11
#define SND_CRASH		12
#define SND_PAUSED		13
#define SND_COPWHISTLE	14
#define SND_GAMEOVER	15



BOOL LoadSound( dword soundno, char *lpszfilename );

BOOL InitDSound(HWND hwnd, GUID *pguid);
BOOL SetupStreamBuffer(LPSTR lpzFileName);
BOOL StreamToBuffer(DWORD dwPos);
void DSExit(void);


// DirectMusic
IDirectMusicPerformance* CreatePerformance(void);
IDirectMusicLoader* CreateLoader(void);
IDirectMusicSegment* LoadSegment(IDirectMusicLoader* pLoader, char *filename);



// Rush Hour specific functions ...

void intro( void );

void showpalette( void );


int drawcitymap( void );
int loadgraphics( void );

int loadcitymap( int mapno );

int newcar( void );
int updatecars( void );
int drawcars( void );

int drawlights( void );
int drawinterface( void );

int editmap( void );
int rungame( void );


int copycitygraphic( void );

int choosefile( BYTE type );


void rollcredits( void );

void showorderinfo( void );
