/*
sound.cpp
sound functions ...

Copyright (C) 1995-2005  Everett Bell

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <windows.h>
#include <stdio.h>
#include <dsound.h>

extern HWND glhwnd;//defined in test2.cpp

#define MAX_SOUNDS 16

LPDIRECTSOUND        glDS = NULL;
LPDIRECTSOUNDBUFFER  glDSB[MAX_SOUNDS];

//wave stuff
typedef struct tagWAVEFILE
{
  DWORD           cbSize;                // Size of file
  LPWAVEFORMATEX  pwfxInfo;              // Wave Header
  LPBYTE          pbData;                // Wave Bits
}
WAVEFILE, *LPWAVEFILE;


//local Function Prototypes
int glInitSound();
void glCloseSound(void);
void glPlaySound(int n,int vol);
void glPlaySoundLooping(int n,int vol);
void glStopSound(int n);
int glLoadWAV(char *fname);

int LoadWave(char *fname,LPDIRECTSOUNDBUFFER &lpDSB);
LPVOID WAVE_Load(char *fname, LPWAVEFILE pWaveFile);
BOOL wave_ParseWaveMemory(void *pvRes, WAVEFORMATEX **ppWaveHeader,
                          BYTE **ppbWaveData, DWORD *pcbWaveSize);



//INIT/CLOSE

int glInitSound()
{
  int a;

  if (DirectSoundCreate(NULL, &glDS, NULL) == DS_OK)
  {
    glDS->SetCooperativeLevel(glhwnd, DSSCL_NORMAL);
    for(a=0; a<MAX_SOUNDS; a++) glDSB[a]=NULL;//clear sound buffers
    return(1);
  }
  else return(0);
}


void glCloseSound(void)
{
  if(glDS!=NULL) glDS->Release();  // this releases buffers as well
}




//PLAY

void glPlaySound(int n,int vol)
{
  if (glDSB[n]!=NULL) {
    glDSB[n]->SetVolume(vol);//volume = 0 to -10,000
    glDSB[n]->Play(0, 0, 0);
  }
}

void glPlaySoundLooping(int n,int vol)
{
  if (glDSB[n]!=NULL) {
    glDSB[n]->SetVolume(vol);//volume = 0 to -10,000
    glDSB[n]->Stop();
  }
}

void glStopSound(int n)
{
  if (glDSB[n]) glDSB[n]->Stop();
}




//LOAD WAV

int glLoadWAV(char *fname)
{
  int a;
  for(a=0; a<MAX_SOUNDS; a++) {
    if (glDSB[a]==NULL) {
      if (!LoadWave(fname, glDSB[a])) return(-1);
      return(a);//return # of snd slot
    }
  }
  return(-1);//no space
}


// LoadWave
//   Gets the sound data and loads it into a DirectSound
//   secondary buffer.

int LoadWave(char *fname, LPDIRECTSOUNDBUFFER &lpDSB)
{
  // These variables are used in Steps 1 and 2
  LPVOID              lpWaveData;
  WAVEFILE            WaveFile;
  DSBUFFERDESC        dsbd;

  // These variables are used in step 3, further down below
  BYTE              *pbData         = NULL;
  BYTE              *pbData2        = NULL;
  DWORD             dwLength;
  DWORD             dwLength2;

  lpWaveData = WAVE_Load(fname, &WaveFile);

  // Step 1: Set up the direct sound buffer.
  memset(&dsbd, 0, sizeof(DSBUFFERDESC));
  dsbd.dwSize                 = sizeof(DSBUFFERDESC);
  // We want a buffer that lives on the sound card's memory
  // (DSBCAPS_STATIC) and can have the pan, volume, and
  // frequency adjusted (DSBCAPS_CTRLDEFAULT)
  dsbd.dwFlags                = DSBCAPS_CTRLDEFAULT | DSBCAPS_STATIC ;

  // Set up the size and format
  dsbd.dwBufferBytes          = WaveFile.cbSize;
  dsbd.lpwfxFormat            = WaveFile.pwfxInfo;  // Must be a PCM format!

  // Step 2: Create the buffer
  if (DS_OK != glDS->CreateSoundBuffer(&dsbd, &lpDSB, NULL)) return(0);

  // Once this code succeeds, lpDSB will point to a DirectSoundBuffer.
  // At this point, you can copy blocks of sound data into the buffer,
  // using the Lock and Unlock interfaces on the DirectSoundBuffer:

  // Lock down the DirectSound buffer
  if (DS_OK == lpDSB->Lock
            (0,               // Offset into buffer to start writing
             WaveFile.cbSize, // Size of wave file to copy in
             (void **)&pbData,         // Points to first block of sound data
             &dwLength,       // Length of first block of data
             (void **)&pbData2,        // Points to second block of sound data
             &dwLength2,      // Length of second block of data
             0L))             // Flags
  {
    // Copy first chunk
    memcpy(pbData, WaveFile.pbData, dwLength);

    // Copy second chunk
    if (dwLength2) memcpy(pbData2, WaveFile.pbData+dwLength , dwLength2);

    // Free up the memory allocated in the WAVE_LoadFile function, since we
    // have copied it to the buffer
    free (lpWaveData);

    // Unlock the buffer
    lpDSB->Unlock(pbData, dwLength, pbData2, dwLength2);

  }
  //else lock failed

  return(1);
}


// WAVE_LoadResource: Gets a wave file into the memory pointed
//                    to by pWaveFile from a resource.

LPVOID WAVE_Load
  (
   char *fname,
   LPWAVEFILE    pWaveFile)  // Points to the struct to fill

{
    LPVOID  lpMemory;

    FILE *fp;
    int length;

    //load wav file into memory
    if ((fp=fopen(fname,"rb"))!=NULL) {
        fseek(fp, 0L, SEEK_END);
        length=ftell(fp);
        if ((lpMemory=malloc(length))==NULL) {fclose(fp); return NULL;}
        fseek(fp, 0L, SEEK_SET);
        fread(lpMemory,1,length,fp);

        // Parse it out
        if (wave_ParseWaveMemory(lpMemory,
                                 &(pWaveFile->pwfxInfo),
                                 &(pWaveFile->pbData),
                                 &(pWaveFile->cbSize)))
          {
            return lpMemory;  // OK
          }
    }

    return NULL;
}


// wave_ParseWaveMemory
//   Parses a chunk of memory into the header and samples.
//   This is done by looking for the "fmt " and "data"
//   fields in the memory.


BOOL wave_ParseWaveMemory

  (LPVOID          lpChunkOfMemory, // Points to raw ram
   LPWAVEFORMATEX  *lplpWaveHeader, // Points to pointer to header
   LPBYTE          *lplpWaveSamples,// Points to pointer to samples
   LPDWORD         lpcbWaveSize)    // Points to size
{
    LPDWORD pdw;
    LPDWORD pdwEnd;
    DWORD   dwRiff;
    DWORD   dwType;
    DWORD   dwLength;

    // Set defaults to NULL or zero
    if (lplpWaveHeader)
        *lplpWaveHeader = NULL;

    if (lplpWaveSamples)
        *lplpWaveSamples = NULL;

    if (lpcbWaveSize)
        *lpcbWaveSize = 0;

    // Set up DWORD pointers to the start of the chunk
    // of memory.
    pdw = (DWORD *)lpChunkOfMemory;

    // Get the type and length of the chunk of memory
    dwRiff = *pdw++;
    dwLength = *pdw++;
    dwType = *pdw++;

    // Using the mmioFOURCC macro (part of Windows SDK), ensure
    // that this is a RIFF WAVE chunk of memory
    if (dwRiff != mmioFOURCC('R', 'I', 'F', 'F'))
      return FALSE;      // not even RIFF

    if (dwType != mmioFOURCC('W', 'A', 'V', 'E'))
      return FALSE;      // not a WAV

    // Find the pointer to the end of the chunk of memory
    pdwEnd = (DWORD *)((BYTE *)pdw + dwLength-4);

    // Run through the bytes looking for the tags
    while (pdw < pdwEnd)
      {
      dwType   = *pdw++;
      dwLength = *pdw++;

      switch (dwType)
        {
        // Found the format part
        case mmioFOURCC('f', 'm', 't', ' '):

          if (lplpWaveHeader && !*lplpWaveHeader)
            {
            if (dwLength < sizeof(WAVEFORMAT))
              return FALSE; // something's wrong! Not a WAV

            // Set the lplpWaveHeader to point to this part of
            // the memory chunk
            *lplpWaveHeader = (LPWAVEFORMATEX)pdw;

            // Check to see if the other two items have been
            // filled out yet (the bits and the size of the
            // bits). If so, then this chunk of memory has
            // been parsed out and we can exit
            if ((!lplpWaveSamples || *lplpWaveSamples) &&
                (!lpcbWaveSize || *lpcbWaveSize))
              {
              return TRUE;
              }
            }
          break;

        // Found the samples
        case mmioFOURCC('d', 'a', 't', 'a'):

          if ((lplpWaveSamples && !*lplpWaveSamples) ||
              (lpcbWaveSize && !*lpcbWaveSize))
            {
            // Point the samples pointer to this part of the
            // chunk of memory.
            if (lplpWaveSamples) *lplpWaveSamples = (LPBYTE)pdw;

            // Set the size of the wave
            if (lpcbWaveSize)    *lpcbWaveSize = dwLength;

            // Make sure we have our header pointer set up.
            // If we do, we can exit
            if (!lplpWaveHeader || *lplpWaveHeader)
              return TRUE;
            }
          break;

        } // End case

      // Move the pointer through the chunk of memory
      pdw = (DWORD *)((BYTE *)pdw + ((dwLength+1)&~1));
      }

  // Failed! If we made it here, we did not get all the pieces
  // of the wave
  return FALSE;
} // wave_ParseWaveMemory



