/*
sprites.cpp
sprite graphics functions ...

Copyright (C) 1995-2005  Everett Bell

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#define WIN32_LEAN_AND_MEAN

#include "globals.h"




void initsprite( byte spriteno, short int width, short int height )
{
        int spritesize;


        if( sprite[spriteno].beingused )
                {
                sprintf(tempstr, "Sprite %d in use!", spriteno);
                cleanup(MISC, tempstr);
                } // end if in use


        sprite[spriteno].height = height;
        sprite[spriteno].width = width;
        sprite[spriteno].visible = 0;
        sprite[spriteno].x = 0;
        sprite[spriteno].y = 0;

        sprite[spriteno].blendamt = 0;

        spritesize = (int)height * (int)width;

//        if( (sprite[spriteno].picptr = (byte *)malloc(spritesize)) == NULL ) cleanup(OUTOFMEM, "initsprite");
//        memoryused += spritesize;
        if( (sprite[spriteno].bgptr = (word *)malloc(spritesize)) == NULL ) cleanup(OUTOFMEM, "initsprite (sprite[].bgptr)");
        memoryused += spritesize * 2;

        sprite[spriteno].beingused = 1;

} // end initsprite



void initspritefrombmp( byte spriteno, char *filename )
{
        int spritesize;


        if( sprite[spriteno].beingused )
                {
                sprintf(tempstr, "Sprite %d in use!", spriteno);
                cleanup(MISC, tempstr);
                } // end if in use

		sprite[spriteno].picptr = loadandunpackbmp( filename );

        sprite[spriteno].height = g_height;
        sprite[spriteno].width = g_width;
        sprite[spriteno].visible = 0;
        sprite[spriteno].x = 0;
        sprite[spriteno].y = 0;

        sprite[spriteno].blendamt = 0;

        spritesize = (int)g_height * (int)g_width * 2;

//        if( (sprite[spriteno].picptr = (byte *)malloc(spritesize)) == NULL ) cleanup(OUTOFMEM, "initsprite");
//        memoryused += spritesize;
        if( (sprite[spriteno].bgptr = (word *)malloc(spritesize)) == NULL ) cleanup(OUTOFMEM, "initsprite (sprite[].bgptr)");
        memoryused += spritesize;

        sprite[spriteno].beingused = 1;

} // end initspritefrombmp




void deletesprite( byte spriteno )
{
        int spritesize;


        if( !(sprite[spriteno].beingused) )
                {
                sprintf( tempstr, "Unable to delete sprite %d ... not in use!", spriteno );
                cleanup( MISC, tempstr );
                }


        spritesize = sprite[spriteno].height * sprite[spriteno].width * 2;

        free(sprite[spriteno].bgptr);
        memoryused -= spritesize;
        free(sprite[spriteno].picptr);
        memoryused -= spritesize;

        sprite[spriteno].beingused = 0;

} // end deletesprite


void showsprite( byte spriteno )
{
        int lineindex, bitindex;
        word *tempscreenptr;
        word *temppicptr;
        word *tempbgptr;

        int spritex, spritey;
        int height, width, widthx2;


        spritex = sprite[spriteno].x;
        spritey = sprite[spriteno].y;
        height = sprite[spriteno].height;
        width = sprite[spriteno].width;
		widthx2 = width<<1;

// Check to keep in boundaries of screen ...
//        if( (spritex<0) || (spritex>639) || (spritey<0) || (spritey>479) ) return;

// TODO Cut off sprites if at edge of screen ...


        temppicptr = sprite[spriteno].picptr;
        tempbgptr = sprite[spriteno].bgptr;

		LOCKVS;

//		tempscreenptr = g_vsptr + VSOFFSET + spritex + (spritey<<9) + (spritey<<8);
		tempscreenptr = g_vsptr + VSOFFSET + spritex + times800(spritey);


		for( lineindex = 0; lineindex < height; lineindex++ )
            {
// Copy whatever was there to bg ...
                memcpy( tempbgptr, tempscreenptr, widthx2 );

                for( bitindex = 0; bitindex < width; bitindex++ )
                {
// TODO -- add shadows here if needed ...
					if( *temppicptr ) *tempscreenptr = *temppicptr;
                    temppicptr++;
                    tempscreenptr++;
                } // end bitindex loop

                tempscreenptr += VSCREENWIDTH-width;
                tempbgptr += width;

            } // end lineindex loop

            sprite[spriteno].visible = 1;

		UNLOCKVS;

} // end showsprite



void showspriteonce( char *tempfilename, int x, int y )
{
		initspritefrombmp( SPRITEONCE, tempfilename );

        sprite[SPRITEONCE].x = x;
        sprite[SPRITEONCE].y = y;
        showsprite( SPRITEONCE );

        deletesprite( SPRITEONCE );

} // end showspriteonce



void showbitmap( char *tempfilename, int x, int y )
{
	LPDIRECTDRAWSURFACE4	lpDDSTemp;

	RECT			rcRect;
//	DDSURFACEDESC2	ddsd;
	HRESULT			ddrval;

	word *tempmem;


// Very BAD way of determining bitmap's height and width ...
// TODO - Fix it!
	tempmem = loadandunpackbmp( tempfilename );
	free( tempmem );
	memoryused -= g_height * g_width * 2;


	lpDDSTemp = DDLoadBitmap( lpDD4, tempfilename, g_width, g_height );
	if( lpDDSTemp == NULL ) cleanup( NOFILE, tempfilename );


//	ZeroMemory( &ddsd, sizeof(ddsd) );
//	ddsd.dwSize = 0;
//	lpDDSTemp->GetSurfaceDesc( &ddsd );


	rcRect.left = 0;
	rcRect.top = 0;
	rcRect.right = g_width;
	rcRect.bottom = g_height;


    while( 1 )
    {
        ddrval = lpDDSVS->BltFast( x, y, lpDDSTemp,
            &rcRect, DDBLTFAST_NOCOLORKEY );
//        ddrval = lpDDSBack->Blt( &rcRect2, lpDDS[CURSORS],
//            &rcRect, DDBLT_KEYSRC, NULL );

		if( ddrval == DD_OK ) break;
		if( ddrval == DDERR_SURFACELOST )
		{
			ddrval = restoreAll();
			if( ddrval != DD_OK )
			{
				ShowDDError( "restoreAll()", ddrval );
				cleanup( OK, NULL );
			}

			ddrval = DDReLoadBitmap( lpDDSTemp, tempfilename );
			if( ddrval != DD_OK )
			{
				ShowDDError( "DDReLoadBitmap() in showbitmap()", ddrval );
				cleanup(OK, NULL );
			} 

			continue;
		}
		if( ddrval != DDERR_WASSTILLDRAWING )
		{
			ShowDDError( "BltFast() in showbitmap()", ddrval );
			MessageBox( hwndmain, tempstr, "Error", MB_OK );
			cleanup( OK, NULL );
		}

	} // end loop

    if(ddrval != DD_OK)
	{
		ShowDDError( "BltFast()", ddrval );
		cleanup( OK, NULL );
	}


	lpDDSTemp->Release();

} // end showbitmap


/*
old way ... before VRAM Surfaces ...
void showbitmap( char *tempfilename, int x, int y )
{
	word *picptr, *temppicptr;
	word *tempscreenptr;

	int index;
	int height, width, widthx2;

	
	picptr = loadandunpackbmp( tempfilename );
	temppicptr = picptr;

	height = g_height;
	width = g_width;
	widthx2 = width<<1;





	LOCKVS;

	tempscreenptr = g_vsptr + VSOFFSET + x + times800(y);

	for( index = 0; index < height; index++ )
	{
		memcpy( tempscreenptr, temppicptr, widthx2 );

		tempscreenptr += VSCREENWIDTH;
		temppicptr += width;
	} // end index loop


	UNLOCKVS;
	
	free( picptr );
	memoryused -= height * widthx2;



} // end showbitmap (derived from showsprite)
*/


void hidesprite( byte spriteno )
{
    int lineindex;

        word *tempbgptr;
        word *tempscreenptr;

        int spritex, spritey;
        int height, width, widthx2;



        spritex = sprite[spriteno].x;
        spritey = sprite[spriteno].y;
        height = sprite[spriteno].height;
        width = sprite[spriteno].width;
		widthx2 = width<<1;

// TODO -- Clip edges of screen ...
// Check to keep in boundaries of screen ...
//        if( (spritex<0) || (spritex>639) || (spritey<0) || (spritey>479) ) return;



// CHANGE!
        tempbgptr = sprite[spriteno].bgptr;

		LOCKVS;

        tempscreenptr = g_vsptr + VSOFFSET + times800(spritey) + spritex;


        for( lineindex = 0; lineindex < height; lineindex++ )
        {
                memcpy( tempscreenptr, tempbgptr, widthx2 );

                tempbgptr += width;
                tempscreenptr += VSCREENWIDTH;

        } // end lineindex loop

		UNLOCKVS;

        sprite[spriteno].visible = 0;

} // end hidesprite


/*
void animatesprites( byte numsprites )
{
        int index;


        for( index = 0; index < numsprites; index++ )
                {
                if( !(animsprite[index].animating) )
                        {
                        if( animsprite[index].frame != 0 )
                                {
                                if( animsprite[index].clockmax < tickcount )
                                        {
                                        redraw = 1;
                                        if( ++(animsprite[index].frame) == animsprite[index].numframes ) animsprite[index].frame = 0;
                                        animsprite[index].clockmax = tickcount + animsprite[index].animspeed;
                                        } // end if time to bump frame

                                } // end if not on frame 0

                        } // end if not animating

                else
                        {
                        if( animsprite[index].clockmax < tickcount )
                                {
                                redraw = 1;
                                if( ++(animsprite[index].frame) == animsprite[index].numframes ) animsprite[index].frame = 0;
                                animsprite[index].clockmax = tickcount + animsprite[index].animspeed;
                                } // end if time to bump frame

                        } // end if animating

                } // end index loop

} // end animatesprites
*/




/*
// 640 x 480 x 8 versions of animsprite functions ...


void initanimsprite( byte spriteno, short int width, short int height, byte numframes,
                      byte speed, byte animating )

{
        int spritesize;


        if( animsprite[spriteno].beingused )
                {
                sprintf(tempstr, "Animated sprite %d in use!", spriteno);
                cleanup(MISC, tempstr);
                } // end if in use


        animsprite[spriteno].height = height;
        animsprite[spriteno].width = width;
        animsprite[spriteno].visible = 0;
        animsprite[spriteno].x = 0;
        animsprite[spriteno].y = 0;

		animsprite[spriteno].blendamt = 0;

        animsprite[spriteno].animating = animating;
        animsprite[spriteno].numframes = numframes;
        animsprite[spriteno].frame = 0;
        animsprite[spriteno].animspeed = speed;

        animsprite[spriteno].clockmax = tickcount + speed;


        spritesize = (int)height * (int)width;

//        if( (animsprite[spriteno].picptr = (byte *)malloc(spritesize)) == NULL ) cleanup(OUTOFMEM, "initanimsprite");
//        memoryused += spritesize;
        if( (animsprite[spriteno].bgptr = (byte *)malloc(spritesize)) == NULL ) cleanup(OUTOFMEM, "initanimsprite");
        memoryused += spritesize;

        animsprite[spriteno].beingused = 1;

} // end initanimsprite


void deleteanimsprite( byte spriteno )
{
        int spritesize;


        spritesize = animsprite[spriteno].height * animsprite[spriteno].width;


        free(animsprite[spriteno].picptr[0]);
        memoryused -= spritesize;

        free(animsprite[spriteno].bgptr);
        memoryused -= spritesize;

        animsprite[spriteno].beingused = 0;

} // end deleteanimsprite


void showanimsprite( byte spriteno )
{
        int lineindex, bitindex;
        byte *tempscreenptr;
        byte *temppicptr;
        byte *tempbgptr;

        short int spritex, spritey;
        short int height, width;
        byte frame;


        spritex = animsprite[spriteno].x;
        spritey = animsprite[spriteno].y;
        height = animsprite[spriteno].height;
        width = animsprite[spriteno].width;
        frame = animsprite[spriteno].frame;

// CHANGE!
        temppicptr = animsprite[spriteno].picptr[frame];
        tempbgptr = animsprite[spriteno].bgptr;

        LOCKVS;
		
		tempscreenptr = g_vsptr + VSOFFSET + (spritey<<9)+(spritey<<7) + spritex;


        for( lineindex = 0; lineindex < height; lineindex++ )
        {

// Copy whatever was there to bg ...
            memcpy( tempbgptr, tempscreenptr, width );

            for( bitindex = 0; bitindex < width; bitindex++ )
            {
                if( *temppicptr ) *tempscreenptr = *temppicptr;
                temppicptr++;
                tempscreenptr++;

            } // end bitindex loop

            tempscreenptr += VSCREENWIDTH-width;
            tempbgptr += width;

        } // end lineindex loop

		UNLOCKVS;

        animsprite[spriteno].visible = 1;

} // end showanimsprite


void hideanimsprite( byte spriteno )
{
		int lineindex;

        byte *tempbgptr;
        byte *tempscreenptr;

        short int spritex, spritey;
        short int height, width;


        spritex = animsprite[spriteno].x;
        spritey = animsprite[spriteno].y;
        height = animsprite[spriteno].height;
        width = animsprite[spriteno].width;


// CHANGE!
        tempbgptr = animsprite[spriteno].bgptr;

		
		LOCKVS;
        
		tempscreenptr = g_vsptr + VSOFFSET + (spritey<<9)+(spritey<<7) + spritex;


        for( lineindex = 0; lineindex < height; lineindex++ )
        {
                memcpy( tempscreenptr, tempbgptr, width<<1 );

                tempbgptr += width;
                tempscreenptr += VSCREENWIDTH;

        } // end lineindex loop


		UNLOCKVS;

        animsprite[spriteno].visible = 0;

} // end hideanimsprite

*/